@@ -1,83 +1,90 @@
+"""XBee API implementation."""
+
 import asyncio
 import binascii
-import enum
 import functools
 import logging
 from typing import Any, Dict, Optional
 
-import serial
 from zigpy.exceptions import APIException, DeliveryError
+import zigpy.types as t
 
-import zigpy_xbee
-from zigpy_xbee.config import CONF_DEVICE_BAUDRATE, CONF_DEVICE_PATH, SCHEMA_DEVICE
+from zigpy_xbee.exceptions import (
+    ATCommandError,
+    ATCommandException,
+    InvalidCommand,
+    InvalidParameter,
+    TransmissionFailure,
+)
 
-from . import types as t, uart
+from . import types as xbee_t, uart
 
 LOGGER = logging.getLogger(__name__)
 
-AT_COMMAND_TIMEOUT = 1
+AT_COMMAND_TIMEOUT = 3
 REMOTE_AT_COMMAND_TIMEOUT = 30
-PROBE_TIMEOUT = 45
-
-
-class ModemStatus(t.uint8_t, t.UndefinedEnum):
-    HARDWARE_RESET = 0x00
-    WATCHDOG_TIMER_RESET = 0x01
-    JOINED_NETWORK = 0x02
-    DISASSOCIATED = 0x03
-    COORDINATOR_STARTED = 0x06
-    NETWORK_SECURITY_KEY_UPDATED = 0x07
-    VOLTAGE_SUPPLY_LIMIT_EXCEEDED = 0x0D
-    MODEM_CONFIGURATION_CHANGED_WHILE_JOIN_IN_PROGRESS = 0x11
-    PAN_ID_CONFLICT_DETECTED = 0x3E
-    PAN_ID_UPDATED_DUE_TO_CONFLICT = 0x3F
-    JOIN_WINDOW_OPENED = 0x43
-    JOIN_WINDOW_CLOSED = 0x44
-    NETWORK_SECURITY_KEY_ROTATION_INITIATED = 0x45
-
-    UNKNOWN_MODEM_STATUS = 0xFF
-    _UNDEFINED = 0xFF
 
 
 # https://www.digi.com/resources/documentation/digidocs/PDFs/90000976.pdf
 COMMAND_REQUESTS = {
-    "at": (0x08, (t.FrameId, t.ATCommand, t.Bytes), 0x88),
-    "queued_at": (0x09, (t.FrameId, t.ATCommand, t.Bytes), 0x88),
+    "at": (0x08, (xbee_t.FrameId, xbee_t.ATCommand, xbee_t.Bytes), 0x88),
+    "queued_at": (0x09, (xbee_t.FrameId, xbee_t.ATCommand, xbee_t.Bytes), 0x88),
     "remote_at": (
         0x17,
-        (t.FrameId, t.EUI64, t.NWK, t.uint8_t, t.ATCommand, t.Bytes),
+        (
+            xbee_t.FrameId,
+            xbee_t.EUI64,
+            xbee_t.NWK,
+            t.uint8_t,
+            xbee_t.ATCommand,
+            xbee_t.Bytes,
+        ),
         0x97,
     ),
     "tx": (0x10, (), None),
     "tx_explicit": (
         0x11,
         (
-            t.FrameId,
-            t.EUI64,
-            t.NWK,
+            xbee_t.FrameId,
+            xbee_t.EUI64,
+            xbee_t.NWK,
             t.uint8_t,
             t.uint8_t,
-            t.uint16_t,
-            t.uint16_t,
+            t.uint16_t_be,
+            t.uint16_t_be,
             t.uint8_t,
             t.uint8_t,
-            t.Bytes,
+            xbee_t.Bytes,
         ),
         0x8B,
     ),
     "create_source_route": (
         0x21,
-        (t.FrameId, t.EUI64, t.NWK, t.uint8_t, t.Relays),
+        (xbee_t.FrameId, xbee_t.EUI64, xbee_t.NWK, t.uint8_t, xbee_t.Relays),
         None,
     ),
-    "register_joining_device": (0x24, (), None),
+    "register_joining_device": (
+        0x24,
+        (xbee_t.FrameId, xbee_t.EUI64, t.uint16_t_be, t.uint8_t, xbee_t.Bytes),
+        0xA4,
+    ),
 }
 COMMAND_RESPONSES = {
-    "at_response": (0x88, (t.FrameId, t.ATCommand, t.uint8_t, t.Bytes), None),
-    "modem_status": (0x8A, (ModemStatus,), None),
+    "at_response": (
+        0x88,
+        (xbee_t.FrameId, xbee_t.ATCommand, t.uint8_t, xbee_t.Bytes),
+        None,
+    ),
+    "modem_status": (0x8A, (xbee_t.ModemStatus,), None),
     "tx_status": (
         0x8B,
-        (t.FrameId, t.NWK, t.uint8_t, t.TXStatus, t.DiscoveryStatus),
+        (
+            xbee_t.FrameId,
+            xbee_t.NWK,
+            t.uint8_t,
+            xbee_t.TXStatus,
+            xbee_t.DiscoveryStatus,
+        ),
         None,
     ),
     "route_information": (0x8D, (), None),
@@ -85,72 +92,85 @@ COMMAND_RESPONSES = {
     "explicit_rx_indicator": (
         0x91,
         (
-            t.EUI64,
-            t.NWK,
+            xbee_t.EUI64,
+            xbee_t.NWK,
             t.uint8_t,
             t.uint8_t,
-            t.uint16_t,
-            t.uint16_t,
+            t.uint16_t_be,
+            t.uint16_t_be,
             t.uint8_t,
-            t.Bytes,
+            xbee_t.Bytes,
         ),
         None,
     ),
     "rx_io_data_long_addr": (0x92, (), None),
     "remote_at_response": (
         0x97,
-        (t.FrameId, t.EUI64, t.NWK, t.ATCommand, t.uint8_t, t.Bytes),
+        (
+            xbee_t.FrameId,
+            xbee_t.EUI64,
+            xbee_t.NWK,
+            xbee_t.ATCommand,
+            t.uint8_t,
+            xbee_t.Bytes,
+        ),
         None,
     ),
     "extended_status": (0x98, (), None),
-    "route_record_indicator": (0xA1, (t.EUI64, t.NWK, t.uint8_t, t.Relays), None),
-    "many_to_one_rri": (0xA3, (t.EUI64, t.NWK, t.uint8_t), None),
+    "route_record_indicator": (
+        0xA1,
+        (xbee_t.EUI64, xbee_t.NWK, t.uint8_t, xbee_t.Relays),
+        None,
+    ),
+    "many_to_one_rri": (0xA3, (xbee_t.EUI64, xbee_t.NWK, t.uint8_t), None),
+    "registration_status": (0xA4, (xbee_t.FrameId, xbee_t.RegistrationStatus), None),
     "node_id_indicator": (0x95, (), None),
 }
 
 # https://www.digi.com/resources/documentation/digidocs/pdfs/90001539.pdf pg 175
 AT_COMMANDS = {
     # Addressing commands
-    "DH": t.uint32_t,
-    "DL": t.uint32_t,
-    "MY": t.uint16_t,
-    "MP": t.uint16_t,
-    "NC": t.uint32_t,  # 0 - MAX_CHILDREN.
-    "SH": t.uint32_t,
-    "SL": t.uint32_t,
+    "DH": t.uint32_t_be,
+    "DL": t.uint32_t_be,
+    "MY": t.uint16_t_be,
+    "MP": t.uint16_t_be,
+    "NC": t.uint32_t_be,  # 0 - MAX_CHILDREN.
+    "SH": t.uint32_t_be,
+    "SL": t.uint32_t_be,
     "NI": t,  # 20 byte printable ascii string
     "SE": t.uint8_t,
     "DE": t.uint8_t,
-    "CI": t.uint16_t,
+    "CI": t.uint16_t_be,
     "TO": t.uint8_t,
-    "NP": t.uint16_t,
-    "DD": t.uint32_t,
+    "NP": t.uint16_t_be,
+    "DD": t.uint32_t_be,
     "CR": t.uint8_t,  # 0 - 0x3F
     # Networking commands
     "CH": t.uint8_t,  # 0x0B - 0x1A
     "DA": t,  # no param
-    "ID": t.uint64_t,
-    "OP": t.uint64_t,
+    "ID": t.uint64_t_be,
+    "OP": t.uint64_t_be,
     "NH": t.uint8_t,
     "BH": t.uint8_t,  # 0 - 0x1E
-    "OI": t.uint16_t,
+    "OI": t.uint16_t_be,
     "NT": t.uint8_t,  # 0x20 - 0xFF
     "NO": t.uint8_t,  # bitfield, 0 - 3
-    "SC": t.uint16_t,  # 1 - 0xFFFF
+    "SC": t.uint16_t_be,  # 1 - 0xFFFF
     "SD": t.uint8_t,  # 0 - 7
     "ZS": t.uint8_t,  # 0 - 2
     "NJ": t.uint8_t,
     "JV": t.Bool,
-    "NW": t.uint16_t,  # 0 - 0x64FF
+    "NW": t.uint16_t_be,  # 0 - 0x64FF
     "JN": t.Bool,
     "AR": t.uint8_t,
     "DJ": t.Bool,  # WTF, docs
-    "II": t.uint16_t,
+    "II": t.uint16_t_be,
     # Security commands
     "EE": t.Bool,
     "EO": t.uint8_t,
-    "NK": t.Bytes,  # 128-bit value
-    "KY": t.Bytes,  # 128-bit value
+    "NK": xbee_t.Bytes,  # 128-bit value
+    "KY": xbee_t.Bytes,  # 128-bit value
+    "KT": t.uint16_t_be,  # 0x1E - 0xFFFF
     # RF interfacing commands
     "PL": t.uint8_t,  # 0 - 4 (basically an Enum)
     "PM": t.Bool,
@@ -166,9 +186,11 @@ AT_COMMANDS = {
     "D7": t.uint8_t,  # 0 - 7 (an Enum)
     "P3": t.uint8_t,  # 0 - 5 (an Enum)
     "P4": t.uint8_t,  # 0 - 5 (an Enum)
+    # MAC diagnostics commands
+    "ED": xbee_t.Bytes,  # 16-byte value
     # I/O commands
-    "IR": t.uint16_t,
-    "IC": t.uint16_t,
+    "IR": t.uint16_t_be,
+    "IC": t.uint16_t_be,
     "D0": t.uint8_t,  # 0 - 5 (an Enum)
     "D1": t.uint8_t,  # 0 - 5 (an Enum)
     "D2": t.uint8_t,  # 0 - 5 (an Enum)
@@ -186,41 +208,40 @@ AT_COMMANDS = {
     "P8": t.uint8_t,  # 0 - 5 (an Enum)
     "P9": t.uint8_t,  # 0 - 5 (an Enum)
     "LT": t.uint8_t,
-    "PR": t.uint16_t,
+    "PR": t.uint16_t_be,
     "RP": t.uint8_t,
-    "%V": t.uint16_t,  # read only
-    "V+": t.uint16_t,
-    "TP": t.uint16_t,
-    "M0": t.uint16_t,  # 0 - 0x3FF
-    "M1": t.uint16_t,  # 0 - 0x3FF
+    "%V": t.uint16_t_be,  # read only
+    "V+": t.uint16_t_be,
+    "TP": t.uint16_t_be,
+    "M0": t.uint16_t_be,  # 0 - 0x3FF
+    "M1": t.uint16_t_be,  # 0 - 0x3FF
     # Diagnostics commands
-    "VR": t.uint16_t,
-    "HV": t.uint16_t,
+    "VR": t.uint16_t_be,
+    "HV": t.uint16_t_be,
     "AI": t.uint8_t,
     # AT command options
-    "CT": t.uint16_t,  # 2 - 0x028F
+    "CT": t.uint16_t_be,  # 2 - 0x028F
     "CN": None,
-    "GT": t.uint16_t,
+    "GT": t.uint16_t_be,
     "CC": t.uint8_t,
     # Sleep commands
     "SM": t.uint8_t,
-    "SN": t.uint16_t,
-    "SP": t.uint16_t,
-    "ST": t.uint16_t,
+    "SN": t.uint16_t_be,
+    "SP": t.uint16_t_be,
+    "ST": t.uint16_t_be,
     "SO": t.uint8_t,
-    "WH": t.uint16_t,
+    "WH": t.uint16_t_be,
     "SI": None,
-    "PO": t.uint16_t,  # 0 - 0x3E8
+    "PO": t.uint16_t_be,  # 0 - 0x3E8
     # Execution commands
     "AC": None,
     "WR": None,
     "RE": None,
     "FR": None,
     "NR": t.Bool,
-    "SI": None,
     "CB": t.uint8_t,
     "ND": t,  # "optional 2-Byte NI value"
-    "DN": t.Bytes,  # "up to 20-Byte printable ASCII string"
+    "DN": xbee_t.Bytes,  # "up to 20-Byte printable ASCII string"
     "IS": None,
     "1S": None,
     "AS": None,
@@ -242,16 +263,19 @@ BAUDRATE_TO_BD = {
 }
 
 
-class ATCommandResult(enum.IntEnum):
-    OK = 0
-    ERROR = 1
-    INVALID_COMMAND = 2
-    INVALID_PARAMETER = 3
-    TX_FAILURE = 4
+AT_COMMAND_RESULT = {
+    1: ATCommandError,
+    2: InvalidCommand,
+    3: InvalidParameter,
+    4: TransmissionFailure,
+}
 
 
 class XBee:
+    """Class implementing XBee communication protocol."""
+
     def __init__(self, device_config: Dict[str, Any]) -> None:
+        """Initialize instance."""
         self._config = device_config
         self._uart: Optional[uart.Gateway] = None
         self._seq: int = 1
@@ -259,7 +283,6 @@ class XBee:
         self._awaiting = {}
         self._app = None
         self._cmd_mode_future: Optional[asyncio.Future] = None
-        self._conn_lost_task: Optional[asyncio.Task] = None
         self._reset: asyncio.Event = asyncio.Event()
         self._running: asyncio.Event = asyncio.Event()
 
@@ -278,80 +301,35 @@ class XBee:
         """Return true if coordinator is running."""
         return self.coordinator_started_event.is_set()
 
-    @classmethod
-    async def new(
-        cls,
-        application: "zigpy_xbee.zigbee.application.ControllerApplication",
-        config: Dict[str, Any],
-    ) -> "XBee":
-        """Create new instance from"""
-        xbee_api = cls(config)
-        await xbee_api.connect()
-        xbee_api.set_application(application)
-        return xbee_api
-
     async def connect(self) -> None:
+        """Connect to the device."""
         assert self._uart is None
         self._uart = await uart.connect(self._config, self)
 
-    def reconnect(self):
-        """Reconnect using saved parameters."""
-        LOGGER.debug(
-            "Reconnecting '%s' serial port using %s",
-            self._config[CONF_DEVICE_PATH],
-            self._config[CONF_DEVICE_BAUDRATE],
-        )
-        return self.connect()
+        try:
+            try:
+                # Ensure we have escaped commands
+                await self._at_command("AP", 2)
+            except asyncio.TimeoutError:
+                if not await self.init_api_mode():
+                    raise APIException("Failed to configure XBee for API mode")
+        except Exception:
+            await self.disconnect()
+            raise
 
     def connection_lost(self, exc: Exception) -> None:
         """Lost serial connection."""
-        LOGGER.warning(
-            "Serial '%s' connection lost unexpectedly: %s",
-            self._config[CONF_DEVICE_PATH],
-            exc,
-        )
-        self._uart = None
-        if self._conn_lost_task and not self._conn_lost_task.done():
-            self._conn_lost_task.cancel()
-        self._conn_lost_task = asyncio.create_task(self._connection_lost())
+        if self._app is not None:
+            self._app.connection_lost(exc)
 
-    async def _connection_lost(self) -> None:
-        """Reconnect serial port."""
-        try:
-            await self._reconnect_till_done()
-        except asyncio.CancelledError:
-            LOGGER.debug("Cancelling reconnection attempt")
-            raise
-
-    async def _reconnect_till_done(self) -> None:
-        attempt = 1
-        while True:
-            try:
-                await asyncio.wait_for(self.reconnect(), timeout=10)
-                break
-            except (asyncio.TimeoutError, OSError) as exc:
-                wait = 2 ** min(attempt, 5)
-                attempt += 1
-                LOGGER.debug(
-                    "Couldn't re-open '%s' serial port, retrying in %ss: %s",
-                    self._config[CONF_DEVICE_PATH],
-                    wait,
-                    str(exc),
-                )
-                await asyncio.sleep(wait)
-
-        LOGGER.debug(
-            "Reconnected '%s' serial port after %s attempts",
-            self._config[CONF_DEVICE_PATH],
-            attempt,
-        )
-
-    def close(self):
+    async def disconnect(self):
+        """Close the connection."""
         if self._uart:
-            self._uart.close()
+            await self._uart.disconnect()
             self._uart = None
 
     def _command(self, name, *args, mask_frame_id=False):
+        """Send API frame to the device."""
         LOGGER.debug("Command %s %s", name, args)
         if self._uart is None:
             raise APIException("API is not running")
@@ -366,6 +344,7 @@ class XBee:
         return future
 
     async def _remote_at_command(self, ieee, nwk, options, name, *args):
+        """Execute AT command on a different XBee module in the network."""
         LOGGER.debug("Remote AT command: %s %s", name, args)
         data = t.serialize(args, (AT_COMMANDS[name],))
         try:
@@ -395,29 +374,30 @@ class XBee:
     _queued_at = functools.partialmethod(_at_partial, "queued_at")
 
     def _api_frame(self, name, *args):
+        """Build API frame."""
         c = COMMAND_REQUESTS[name]
         return (bytes([c[0]]) + t.serialize(args, c[1])), c[2]
 
     def frame_received(self, data):
+        """Handle API frame from the device."""
         command = self._commands_by_id[data[0]]
         LOGGER.debug("Frame received: %s", command)
         data, rest = t.deserialize(data[1:], COMMAND_RESPONSES[command][1])
         try:
-            getattr(self, "_handle_%s" % (command,))(*data)
+            getattr(self, f"_handle_{command}")(*data)
         except AttributeError:
             LOGGER.error("No '%s' handler. Data: %s", command, binascii.hexlify(data))
 
     def _handle_at_response(self, frame_id, cmd, status, value):
+        """Local AT command response."""
         (fut,) = self._awaiting.pop(frame_id)
-        try:
-            status = ATCommandResult(status)
-        except ValueError:
-            status = ATCommandResult.ERROR
 
         if status:
-            fut.set_exception(
-                RuntimeError("AT Command response: {}".format(status.name))
-            )
+            try:
+                exception = AT_COMMAND_RESULT[status]
+            except KeyError:
+                exception = ATCommandException
+            fut.set_exception(exception(f"AT Command response: {status}"))
             return
 
         response_type = AT_COMMANDS[cmd.decode("ascii")]
@@ -442,12 +422,15 @@ class XBee:
     def _handle_modem_status(self, status):
         LOGGER.debug("Handle modem status frame: %s", status)
         status = status
-        if status == ModemStatus.COORDINATOR_STARTED:
+        if status == xbee_t.ModemStatus.COORDINATOR_STARTED:
             self.coordinator_started_event.set()
-        elif status in (ModemStatus.HARDWARE_RESET, ModemStatus.WATCHDOG_TIMER_RESET):
+        elif status in (
+            xbee_t.ModemStatus.HARDWARE_RESET,
+            xbee_t.ModemStatus.WATCHDOG_TIMER_RESET,
+        ):
             self.reset_event.set()
             self.coordinator_started_event.clear()
-        elif status == ModemStatus.DISASSOCIATED:
+        elif status == xbee_t.ModemStatus.DISASSOCIATED:
             self.coordinator_started_event.clear()
 
         if self._app:
@@ -486,17 +469,27 @@ class XBee:
 
         try:
             if tx_status in (
-                t.TXStatus.BROADCAST_APS_TX_ATTEMPT,
-                t.TXStatus.SELF_ADDRESSED,
-                t.TXStatus.SUCCESS,
+                xbee_t.TXStatus.BROADCAST_APS_TX_ATTEMPT,
+                xbee_t.TXStatus.SELF_ADDRESSED,
+                xbee_t.TXStatus.SUCCESS,
             ):
                 fut.set_result(tx_status)
             else:
-                fut.set_exception(DeliveryError("%s" % (tx_status,)))
+                fut.set_exception(DeliveryError(f"{tx_status}"))
         except asyncio.InvalidStateError as ex:
             LOGGER.debug("duplicate tx_status for %s nwk? State: %s", nwk, ex)
 
+    def _handle_registration_status(self, frame_id, status):
+        (fut,) = self._awaiting.pop(frame_id)
+        if status:
+            fut.set_exception(RuntimeError(f"Registration Status: {status.name}"))
+            return
+        LOGGER.debug(f"Registration Status: {status.name}")
+
+        fut.set_result(status)
+
     def set_application(self, app):
+        """Set reference to ControllerApplication."""
         self._app = app
 
     def handle_command_mode_rsp(self, data):
@@ -512,7 +505,7 @@ class XBee:
             fut.set_result(data)
 
     async def command_mode_at_cmd(self, command):
-        """Sends AT command in command mode."""
+        """Send AT command in command mode."""
         self._cmd_mode_future = asyncio.Future()
         self._uart.command_mode_send(command.encode("ascii"))
 
@@ -540,7 +533,7 @@ class XBee:
             if not await self.command_mode_at_cmd(cmd + "\r"):
                 LOGGER.debug("No response to %s cmd", cmd)
                 return None
-            LOGGER.debug("Successfuly sent %s cmd", cmd)
+            LOGGER.debug("Successfully sent %s cmd", cmd)
         self._uart.reset_command_mode()
         return True
 
@@ -565,44 +558,13 @@ class XBee:
                 return res
 
         LOGGER.debug(
-            (
-                "Couldn't enter AT command mode at any known baudrate."
-                "Configure XBee manually for escaped API mode ATAP2"
-            )
+            "Couldn't enter AT command mode at any known baudrate."
+            "Configure XBee manually for escaped API mode ATAP2"
         )
         return False
 
-    @classmethod
-    async def probe(cls, device_config: Dict[str, Any]) -> bool:
-        """Probe port for the device presence."""
-        api = cls(SCHEMA_DEVICE(device_config))
-        try:
-            await asyncio.wait_for(api._probe(), timeout=PROBE_TIMEOUT)
-            return True
-        except (asyncio.TimeoutError, serial.SerialException, APIException) as exc:
-            LOGGER.debug(
-                "Unsuccessful radio probe of '%s' port",
-                device_config[CONF_DEVICE_PATH],
-                exc_info=exc,
-            )
-        finally:
-            api.close()
-
-        return False
-
-    async def _probe(self) -> None:
-        """Open port and try sending a command"""
-        await self.connect()
-        try:
-            # Ensure we have escaped commands
-            await self._at_command("AP", 2)
-        except asyncio.TimeoutError:
-            if not await self.init_api_mode():
-                raise APIException("Failed to configure XBee for API mode")
-        finally:
-            self.close()
-
     def __getattr__(self, item):
+        """Handle supported command requests."""
         if item in COMMAND_REQUESTS:
             return functools.partial(self._command, item)
-        raise AttributeError("Unknown command {}".format(item))
+        raise AttributeError(f"Unknown command {item}")
