@@ -1,15 +1,18 @@
+"""Module for UART communication to the device."""
+
 import asyncio
 import logging
 from typing import Any, Dict
 
+import zigpy.config
 import zigpy.serial
 
-from zigpy_xbee.config import CONF_DEVICE_BAUDRATE, CONF_DEVICE_PATH
-
 LOGGER = logging.getLogger(__name__)
 
 
-class Gateway(asyncio.Protocol):
+class Gateway(zigpy.serial.SerialProtocol):
+    """Class implementing the UART protocol."""
+
     START = b"\x7E"
     ESCAPE = b"\x7D"
     XON = b"\x11"
@@ -18,14 +21,14 @@ class Gateway(asyncio.Protocol):
     RESERVED = START + ESCAPE + XON + XOFF
     THIS_ONE = True
 
-    def __init__(self, api, connected_future=None):
-        self._buffer = b""
-        self._connected_future = connected_future
+    def __init__(self, api):
+        """Initialize instance."""
+        super().__init__()
         self._api = api
         self._in_command_mode = False
 
     def send(self, data):
-        """Send data, taking care of escaping and framing"""
+        """Send data, taking care of escaping and framing."""
         LOGGER.debug("Sending: %s", data)
         checksum = bytes([self._checksum(data)])
         frame = self.START + self._escape(
@@ -45,32 +48,18 @@ class Gateway(asyncio.Protocol):
             self._transport.serial.baudrate = baudrate
         else:
             raise ValueError(
-                "baudrate must be one of {}".format(self._transport.serial.BAUDRATES)
+                f"baudrate must be one of {self._transport.serial.BAUDRATES}"
             )
 
     def connection_lost(self, exc) -> None:
         """Port was closed expectedly or unexpectedly."""
-        if self._connected_future and not self._connected_future.done():
-            if exc is None:
-                self._connected_future.set_result(True)
-            else:
-                self._connected_future.set_exception(exc)
-        if exc is None:
-            LOGGER.debug("Closed serial connection")
-            return
+        super().connection_lost(exc)
 
-        LOGGER.error("Lost serial connection: %s", exc)
-        self._api.connection_lost(exc)
-
-    def connection_made(self, transport):
-        """Callback when the uart is connected"""
-        LOGGER.debug("Connection made")
-        self._transport = transport
-        if self._connected_future:
-            self._connected_future.set_result(True)
+        if self._api is not None:
+            self._api.connection_lost(exc)
 
     def command_mode_rsp(self, data):
-        """Handles AT command mode response."""
+        """Handle AT command mode response."""
         data = data.decode("ascii", "ignore")
         LOGGER.debug("Handling AT command mode response: %s", data)
         self._api.handle_command_mode_rsp(data)
@@ -82,27 +71,25 @@ class Gateway(asyncio.Protocol):
         self._transport.write(data)
 
     def data_received(self, data):
-        """Callback when there is data received from the uart"""
-        self._buffer += data
+        """Handle data received from the UART callback."""
+        super().data_received(data)
         while self._buffer:
             frame = self._extract_frame()
             if frame is None:
                 break
             self.frame_received(frame)
         if self._in_command_mode and self._buffer[-1:] == b"\r":
-            rsp, self._buffer = (self._buffer[:-1], b"")
+            rsp = self._buffer[:-1]
+            self._buffer.clear()
             self.command_mode_rsp(rsp)
 
     def frame_received(self, frame):
-        """Frame receive handler"""
+        """Frame receive handler."""
         LOGGER.debug("Frame received: %s", frame)
         self._api.frame_received(frame)
 
-    def close(self):
-        self._transport.close()
-
     def reset_command_mode(self):
-        """Reset command mode and ignore \r character as command mode response."""
+        r"""Reset command mode and ignore '\r' character as command mode response."""
         self._in_command_mode = False
 
     def _extract_frame(self):
@@ -161,21 +148,16 @@ class Gateway(asyncio.Protocol):
         return 0xFF - (sum(data) % 0x100)
 
 
-async def connect(device_config: Dict[str, Any], api, loop=None) -> Gateway:
-    if loop is None:
-        loop = asyncio.get_event_loop()
-
-    connected_future = asyncio.Future()
-    protocol = Gateway(api, connected_future)
-
+async def connect(device_config: Dict[str, Any], api) -> Gateway:
+    """Connect to the device."""
     transport, protocol = await zigpy.serial.create_serial_connection(
-        loop,
-        lambda: protocol,
-        url=device_config[CONF_DEVICE_PATH],
-        baudrate=device_config[CONF_DEVICE_BAUDRATE],
-        xonxoff=False,
+        loop=asyncio.get_running_loop(),
+        protocol_factory=lambda: Gateway(api),
+        url=device_config[zigpy.config.CONF_DEVICE_PATH],
+        baudrate=device_config[zigpy.config.CONF_DEVICE_BAUDRATE],
+        xonxoff=device_config[zigpy.config.CONF_DEVICE_BAUDRATE],
     )
 
-    await connected_future
+    await protocol.wait_until_connected()
 
     return protocol
