@@ -1,15 +1,20 @@
+"""ControllerApplication for XBee adapters."""
+
 from __future__ import annotations
 
 import asyncio
 import logging
-import time
+import math
+import statistics
 from typing import Any
 
 import zigpy.application
 import zigpy.config
+from zigpy.config import CONF_DEVICE
 import zigpy.device
 import zigpy.exceptions
 import zigpy.quirks
+import zigpy.state
 import zigpy.types
 import zigpy.util
 from zigpy.zcl import foundation
@@ -18,7 +23,8 @@ import zigpy.zdo.types as zdo_t
 
 import zigpy_xbee
 import zigpy_xbee.api
-from zigpy_xbee.config import CONF_DEVICE, CONFIG_SCHEMA, SCHEMA_DEVICE
+import zigpy_xbee.config
+from zigpy_xbee.exceptions import InvalidCommand
 from zigpy_xbee.types import EUI64, UNKNOWN_IEEE, UNKNOWN_NWK, TXOptions, TXStatus
 
 # how long coordinator would hold message for an end device in 10ms units
@@ -35,33 +41,32 @@ XBEE_ENDPOINT_ID = 0xE6
 
 
 class ControllerApplication(zigpy.application.ControllerApplication):
-    SCHEMA = CONFIG_SCHEMA
-    SCHEMA_DEVICE = SCHEMA_DEVICE
+    """Implementation of Zigpy ControllerApplication for XBee devices."""
 
-    probe = zigpy_xbee.api.XBee.probe
+    CONFIG_SCHEMA = zigpy_xbee.config.CONFIG_SCHEMA
 
     def __init__(self, config: dict[str, Any]):
-        super().__init__(config=zigpy.config.ZIGPY_SCHEMA(config))
+        """Initialize instance."""
+        super().__init__(config=config)
         self._api: zigpy_xbee.api.XBee | None = None
+        self.topology.add_listener(self)
 
     async def disconnect(self):
         """Shutdown application."""
         if self._api:
-            self._api.close()
+            await self._api.disconnect()
+            self._api = None
 
     async def connect(self):
-        self._api = await zigpy_xbee.api.XBee.new(self, self._config[CONF_DEVICE])
-        try:
-            # Ensure we have escaped commands
-            await self._api._at_command("AP", 2)
-        except asyncio.TimeoutError:
-            LOGGER.debug("No response to API frame. Configure API mode")
-            if not await self._api.init_api_mode():
-                raise zigpy.exceptions.ControllerException(
-                    "Failed to configure XBee API mode."
-                )
+        """Connect to the device."""
+        api = zigpy_xbee.api.XBee(self._config[CONF_DEVICE])
+        await api.connect()
+        api.set_application(self)
+
+        self._api = api
 
     async def start_network(self):
+        """Configure the module to work with Zigpy."""
         association_state = await asyncio.wait_for(
             self._get_association_state(), timeout=4
         )
@@ -69,13 +74,16 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         # Enable ZDO passthrough
         await self._api._at_command("AO", 0x03)
 
+        if self.state.node_info == zigpy.state.NodeInfo():
+            await self.load_network_info()
+
         enc_enabled = await self._api._at_command("EE")
         enc_options = await self._api._at_command("EO")
         zb_profile = await self._api._at_command("ZS")
 
         if (
             enc_enabled != 1
-            or enc_options != 2
+            or enc_options & 0b0010 != 0b0010
             or zb_profile != 2
             or association_state != 0
             or self.state.node_info.nwk != 0x0000
@@ -99,7 +107,10 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         self.listener_event("raw_device_initialized", xbee_dev)
         self.devices[dev.ieee] = xbee_dev
 
+        await self.register_endpoints()
+
     async def load_network_info(self, *, load_devices=False):
+        """Load supported parameters of network_info and node_info from the device."""
         # Load node info
         node_info = self.state.node_info
         node_info.nwk = zigpy.types.NWK(await self._api._at_command("MY"))
@@ -114,10 +125,17 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                 node_info.logical_type = zdo_t.LogicalType.Coordinator
             else:
                 node_info.logical_type = zdo_t.LogicalType.EndDevice
-        except RuntimeError:
+        except InvalidCommand:
             LOGGER.warning("CE command failed, assuming node is coordinator")
             node_info.logical_type = zdo_t.LogicalType.Coordinator
 
+        # TODO: Feature detect the XBee's exact model
+        node_info.model = "XBee"
+        node_info.manufacturer = "Digi"
+
+        version = await self._api._at_command("VR")
+        node_info.version = f"{int(version):#06x}"
+
         # Load network info
         pan_id = await self._api._at_command("OI")
         extended_pan_id = await self._api._at_command("ID")
@@ -131,26 +149,30 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         network_info.channel = await self._api._at_command("CH")
 
     async def reset_network_info(self) -> None:
+        """Reset Zigbee network."""
         await self._api._at_command("NR", 0)
 
     async def write_network_info(self, *, network_info, node_info):
-        scan_bitmask = 1 << (network_info.channel - 11)
+        """Write supported network_info and node_info parameters to the device."""
+        epid, _ = zigpy.types.uint64_t.deserialize(
+            network_info.extended_pan_id.serialize()
+        )
+        await self._api._queued_at("ID", epid)
 
         await self._api._queued_at("ZS", 2)
+        scan_bitmask = 1 << (network_info.channel - 11)
         await self._api._queued_at("SC", scan_bitmask)
         await self._api._queued_at("EE", 1)
-        await self._api._queued_at("EO", 2)
-
+        await self._api._queued_at("EO", 0b0010)
         await self._api._queued_at("NK", network_info.network_key.key.serialize())
         await self._api._queued_at("KY", network_info.tc_link_key.key.serialize())
-
         await self._api._queued_at("NJ", 0)
         await self._api._queued_at("SP", CONF_CYCLIC_SLEEP_PERIOD)
         await self._api._queued_at("SN", CONF_POLL_TIMEOUT)
 
         try:
             await self._api._queued_at("CE", 1)
-        except RuntimeError:
+        except InvalidCommand:
             pass
 
         await self._api._at_command("WR")
@@ -161,14 +183,67 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         )
         LOGGER.debug("Association state: %s", association_state)
 
+    async def _move_network_to_channel(
+        self, new_channel: int, new_nwk_update_id: int
+    ) -> None:
+        """Move the coordinator to a new channel."""
+        scan_bitmask = 1 << (new_channel - 11)
+        await self._api._queued_at("SC", scan_bitmask)
+
+    async def energy_scan(
+        self, channels: zigpy.types.Channels, duration_exp: int, count: int
+    ) -> dict[int, float]:
+        """Run an energy detection scan and returns the per-channel scan results."""
+        all_results = {}
+
+        for _ in range(count):
+            try:
+                results = await self._api._at_command("ED", bytes([duration_exp]))
+            except InvalidCommand:
+                LOGGER.warning("Coordinator does not support energy scanning")
+                return {c: 0 for c in channels}
+
+            results = {
+                channel: -int(rssi) for channel, rssi in zip(range(11, 27), results)
+            }
+
+            for channel, rssi in results.items():
+                all_results.setdefault(channel, []).append(rssi)
+
+        def logistic(x: float, *, L: float = 1, x_0: float = 0, k: float = 1) -> float:
+            """Logistic function."""
+            return L / (1 + math.exp(-k * (x - x_0)))
+
+        def map_rssi_to_energy(rssi: int) -> float:
+            """Remaps RSSI (in dBm) to Energy (0-255)."""
+            RSSI_MAX = -5
+            RSSI_MIN = -92
+            return logistic(
+                x=rssi,
+                L=255,
+                x_0=RSSI_MIN + 0.45 * (RSSI_MAX - RSSI_MIN),
+                k=0.13,
+            )
+
+        energy = {
+            channel: map_rssi_to_energy(statistics.mean(all_rssi))
+            for channel, all_rssi in all_results.items()
+        }
+
+        return {channel: energy.get(channel, 0) for channel in channels}
+
     async def force_remove(self, dev):
         """Forcibly remove device from NCP."""
-        pass
 
-    async def add_endpoint(self, descriptor):
+    async def add_endpoint(self, descriptor: zdo_t.SimpleDescriptor) -> None:
         """Register a new endpoint on the device."""
-        # This is not provided by the XBee API
-        pass
+        self._device.replacement["endpoints"][descriptor.endpoint] = {
+            "device_type": descriptor.device_type,
+            "profile_id": descriptor.profile,
+            "input_clusters": descriptor.input_clusters,
+            "output_clusters": descriptor.output_clusters,
+        }
+        self._device.add_endpoint(descriptor.endpoint)
 
     async def _get_association_state(self):
         """Wait for Zigbee to start."""
@@ -180,6 +255,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         return state
 
     async def send_packet(self, packet: zigpy.types.ZigbeePacket) -> None:
+        """Send ZigbeePacket via the device."""
         LOGGER.debug("Sending packet %r", packet)
 
         try:
@@ -242,10 +318,11 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                 f"Failed to deliver packet: {v!r}", status=v
             )
 
-    @zigpy.util.retryable_request
+    @zigpy.util.retryable_request()
     def remote_at_command(
         self, nwk, cmd_name, *args, apply_changes=True, encryption=True
     ):
+        """Execute AT command on another XBee module in the network."""
         LOGGER.debug("Remote AT%s command: %s", cmd_name, args)
         options = zigpy.types.uint8_t(0)
         if apply_changes:
@@ -256,69 +333,111 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         return self._api._remote_at_command(dev.ieee, nwk, options, cmd_name, *args)
 
     async def permit_ncp(self, time_s=60):
+        """Permit join."""
         assert 0 <= time_s <= 254
         await self._api._at_command("NJ", time_s)
         await self._api._at_command("AC")
 
-    async def permit_with_key(self, node, code, time_s=60):
-        raise NotImplementedError("XBee does not support install codes")
+    async def permit_with_link_key(
+        self, node: EUI64, link_key: zigpy.types.KeyData, time_s: int = 500, key_type=0
+    ):
+        """Permits a new device to join with the given IEEE and link key."""
+        assert 0x1E <= time_s <= 0xFFFF
+        await self._api._at_command("KT", time_s)
+        reserved = 0xFFFE
+        # Key type:
+        # 0 = Pre-configured Link Key (KY command of the joining device)
+        # 1 = Install Code With CRC (I? command of the joining device)
+        await self._api.register_joining_device(node, reserved, key_type, link_key)
 
     def handle_modem_status(self, status):
+        """Handle changed Modem Status of the device."""
         LOGGER.info("Modem status update: %s (%s)", status.name, status.value)
 
     def handle_rx(
         self, src_ieee, src_nwk, src_ep, dst_ep, cluster_id, profile_id, rxopts, data
     ):
-        if src_nwk == 0:
+        """Handle receipt of Zigbee data from the device."""
+        src = zigpy.types.AddrModeAddress(
+            addr_mode=zigpy.types.AddrMode.NWK, address=src_nwk
+        )
+
+        dst = zigpy.types.AddrModeAddress(
+            addr_mode=zigpy.types.AddrMode.NWK, address=self.state.node_info.nwk
+        )
+
+        if src == dst:
             LOGGER.info("handle_rx self addressed")
 
-        ember_ieee = zigpy.types.EUI64(src_ieee)
-        if dst_ep == 0 and cluster_id == zdo_t.ZDOCmd.Device_annce:
-            # ZDO Device announce request
-            nwk, rest = zigpy.types.NWK.deserialize(data[1:])
-            ieee, rest = zigpy.types.EUI64.deserialize(rest)
-            LOGGER.info("New device joined: NWK 0x%04x, IEEE %s", nwk, ieee)
-            if ember_ieee != ieee:
-                LOGGER.warning(
-                    "Announced IEEE %s is different from originator %s",
-                    str(ieee),
-                    str(ember_ieee),
-                )
-            if src_nwk != nwk:
-                LOGGER.warning(
-                    "Announced 0x%04x NWK is different from originator 0x%04x",
-                    nwk,
-                    src_nwk,
-                )
-            self.handle_join(nwk, ieee, 0)
-
         try:
-            self._device.last_seen = time.time()
+            self._device.update_last_seen()
         except KeyError:
             pass
 
-        try:
-            device = self.get_device(nwk=src_nwk)
-        except KeyError:
-            if ember_ieee != UNKNOWN_IEEE and ember_ieee in self.devices:
-                self.handle_join(src_nwk, ember_ieee, 0)
-                device = self.get_device(ieee=ember_ieee)
-            else:
-                LOGGER.debug(
-                    "Received frame from unknown device: 0x%04x/%s",
-                    src_nwk,
-                    str(ember_ieee),
-                )
-                return
+        self.packet_received(
+            zigpy.types.ZigbeePacket(
+                src=src,
+                src_ep=src_ep,
+                dst=dst,
+                dst_ep=dst_ep,
+                tsn=None,
+                profile_id=profile_id,
+                cluster_id=cluster_id,
+                data=zigpy.types.SerializableBytes(data),
+            )
+        )
 
-        self.handle_message(device, profile_id, cluster_id, src_ep, dst_ep, data)
+    def neighbors_updated(
+        self, ieee: zigpy.types.EUI64, neighbors: list[zdo_t.Neighbor]
+    ) -> None:
+        """Neighbor update from Mgmt_Lqi_req."""
+        for neighbor in neighbors:
+            if neighbor.relationship == zdo_t.Neighbor.Relationship.Parent:
+                device = self.get_device(ieee=ieee)
+                device.radio_details(lqi=neighbor.lqi)
+
+            elif neighbor.relationship == zdo_t.Neighbor.Relationship.Child:
+                try:
+                    child_device = self.get_device(ieee=neighbor.ieee)
+                    child_device.radio_details(lqi=neighbor.lqi)
+                except KeyError:
+                    LOGGER.warning("Unknown device %r", neighbor.ieee)
+
+    def routes_updated(
+        self, ieee: zigpy.types.EUI64, routes: list[zdo_t.Route]
+    ) -> None:
+        """Route update from Mgmt_Rtg_req."""
+        self.create_task(
+            self._routes_updated(ieee, routes), f"routes_updated-ieee={ieee}"
+        )
+
+    async def _routes_updated(
+        self, ieee: zigpy.types.EUI64, routes: list[zdo_t.Route]
+    ) -> None:
+        """Get RSSI for adjacent routers on Route update from Mgmt_Rtg_req."""
+        for route in routes:
+            if (
+                route.DstNWK == self.state.node_info.nwk
+                and route.NextHop == self.state.node_info.nwk
+                and route.RouteStatus == zdo_t.RouteStatus.Active
+            ):
+                device = self.get_device(ieee=ieee)
+                rssi = await self._api._at_command("DB")
+                device.radio_details(rssi=-rssi)
+                break
 
 
 class XBeeCoordinator(zigpy.quirks.CustomDevice):
+    """Zigpy Device representing Coordinator."""
+
     class XBeeGroup(zigpy.quirks.CustomCluster, Groups):
+        """XBeeGroup custom cluster."""
+
         cluster_id = 0x0006
 
     class XBeeGroupResponse(zigpy.quirks.CustomCluster, Groups):
+        """XBeeGroupResponse custom cluster."""
+
         cluster_id = 0x8006
         ep_attribute = "xbee_groups_response"
 
@@ -332,6 +451,8 @@ class XBeeCoordinator(zigpy.quirks.CustomDevice):
         }
 
     def __init__(self, *args, **kwargs):
+        """Initialize instance."""
+
         super().__init__(*args, **kwargs)
         self.node_desc = zdo_t.NodeDescriptor(
             logical_type=zdo_t.LogicalType.Coordinator,
